"""A window with a textfield where the user can edit.

Useful for editing the contents of an article.

.. note:: idlelib, tkinter and pillow modules are required.

.. warning::
   With Pillow 10.0, 10.1 no wheels for 32-bit Python on Windows are
   supported. Pillow 10.2 supports it again. Either you have to update
   your Python using a 64-bit version or you have to
   :command:`pip install "pillow>8.1.1,!=10.0,!=10.1"`.

.. seealso:: :mod:`editor`
"""
#
# (C) Pywikibot team, 2003-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot


# Some Python distributions have tkinter but the underlying _tkinter
# implementation is missing. Thus just import tkinter does not raise
# the exception. Therefore try to import _tkinter.
# Note: idlelib also needs tkinter.
try:
    import _tkinter  # noqa: F401
except ImportError as e:
    idlelib = tkinter = e
    Frame = simpledialog = ScrolledText = object
    ConfigDialog = ReplaceDialog = SearchDialog = object
    idleConf = MultiCallCreator = object  # noqa:  N816
else:
    import idlelib
    import tkinter
    from idlelib import replace as ReplaceDialog  # noqa: N812
    from idlelib import search as SearchDialog  # noqa: N812
    from idlelib.config import idleConf
    from idlelib.configdialog import ConfigDialog
    from idlelib.multicall import MultiCallCreator
    from tkinter import Frame, simpledialog
    from tkinter.scrolledtext import ScrolledText


__all__ = ('EditBoxWindow', 'TextEditor', 'Tkdialog')


class TextEditor(ScrolledText):

    """A text widget with some editing enhancements.

    A lot of code here is copied or adapted from the
    idlelib/EditorWindow.py file in the standard Python distribution.
    """

    def __init__(self, master=None, **kwargs) -> None:
        """Initializer.

        Get default settings from user's IDLE configuration.
        """
        for module in (idlelib, tkinter):
            if isinstance(module, ImportError):
                raise module

        textcf = self._initialize_config(idleConf.CurrentTheme())

        if idleConf.GetOption('main', 'EditorWindow', 'font-bold',
                              type='bool'):
            font_weight = 'bold'
        else:
            font_weight = 'normal'
        textcf['font'] = (
            idleConf.GetOption('main', 'EditorWindow', 'font'),
            idleConf.GetOption('main', 'EditorWindow', 'font-size'),
            font_weight)

        # override defaults with any user-specified settings
        textcf.update(kwargs)
        super().__init__(master, **textcf)

    @staticmethod
    def _initialize_config(theme):
        """Fix idleConf.GetHighlight method for different Python releases."""
        return {
            'padx': 5,
            'wrap': 'word',
            'undo': 'True',
            'width': idleConf.GetOption('main', 'EditorWindow', 'width'),
            'height': idleConf.GetOption('main', 'EditorWindow', 'height'),
            'foreground': idleConf.GetHighlight(theme, 'normal')['foreground'],
            'background': idleConf.GetHighlight(theme, 'normal')['background'],
            'highlightcolor': idleConf.GetHighlight(
                theme, 'hilite')['foreground'],
            'highlightbackground': idleConf.GetHighlight(
                theme, 'hilite')['background'],
            'insertbackground': idleConf.GetHighlight(
                theme, 'cursor')['foreground'],
        }

    def add_bindings(self) -> None:
        """Assign key and events bindings to methods."""
        # due to IDLE dependencies, this can't be called from __init__
        # add key and event bindings
        self.bind('<<cut>>', self.cut)
        self.bind('<<copy>>', self.copy)
        self.bind('<<paste>>', self.paste)
        self.bind('<<select-all>>', self.select_all)
        self.bind('<<remove-selection>>', self.remove_selection)
        self.bind('<<find>>', self.find_event)
        self.bind('<<find-again>>', self.find_again_event)
        self.bind('<<find-selection>>', self.find_selection_event)
        self.bind('<<replace>>', self.replace_event)
        self.bind('<<goto-line>>', self.goto_line_event)
        self.bind('<<del-word-left>>', self.del_word_left)
        self.bind('<<del-word-right>>', self.del_word_right)
        keydefs = {'<<copy>>': ['<Control-Key-c>', '<Control-Key-C>'],
                   '<<cut>>': ['<Control-Key-x>', '<Control-Key-X>'],
                   '<<del-word-left>>': ['<Control-Key-BackSpace>'],
                   '<<del-word-right>>': ['<Control-Key-Delete>'],
                   '<<end-of-file>>': ['<Control-Key-d>', '<Control-Key-D>'],
                   '<<find-again>>': ['<Control-Key-g>', '<Key-F3>'],
                   '<<find-selection>>': ['<Control-Key-F3>'],
                   '<<find>>': ['<Control-Key-f>', '<Control-Key-F>'],
                   '<<goto-line>>': ['<Alt-Key-g>', '<Meta-Key-g>'],
                   '<<paste>>': ['<Control-Key-v>', '<Control-Key-V>'],
                   '<<redo>>': ['<Control-Shift-Key-Z>'],
                   '<<remove-selection>>': ['<Key-Escape>'],
                   '<<replace>>': ['<Control-Key-h>', '<Control-Key-H>'],
                   '<<select-all>>': ['<Control-Key-a>'],
                   '<<undo>>': ['<Control-Key-z>', '<Control-Key-Z>'],
                   }

        for event, keylist in keydefs.items():
            if keylist:
                self.event_add(event, *keylist)

    def cut(self, event) -> str:
        """Perform cut operation."""
        if self.tag_ranges('sel'):
            self.event_generate('<<Cut>>')
        return 'break'

    def copy(self, event) -> str:
        """Perform copy operation."""
        if self.tag_ranges('sel'):
            self.event_generate('<<Copy>>')
        return 'break'

    def paste(self, event) -> str:
        """Perform paste operation."""
        self.event_generate('<<Paste>>')
        return 'break'

    def select_all(self, event=None) -> str:
        """Perform select all operation."""
        self.tag_add('sel', '1.0', 'end-1c')
        self.mark_set('insert', '1.0')
        self.see('insert')
        return 'break'

    def remove_selection(self, event=None) -> None:
        """Perform remove operation."""
        self.tag_remove('sel', '1.0', 'end')
        self.see('insert')

    def del_word_left(self, event) -> str:
        """Perform delete word (left) operation."""
        self.event_generate('<Meta-Delete>')
        return 'break'

    def del_word_right(self, event=None) -> str:
        """Perform delete word (right) operation."""
        self.event_generate('<Meta-d>')
        return 'break'

    def find_event(self, event=None) -> str:
        """Perform find operation."""
        if not self.tag_ranges('sel'):
            found = self.tag_ranges('found')
            if found:
                self.tag_add('sel', found[0], found[1])
            else:
                self.tag_add('sel', '1.0', '1.0+1c')
        SearchDialog.find(self)
        return 'break'

    def find_again_event(self, event=None) -> str:
        """Perform find again operation."""
        SearchDialog.find_again(self)
        return 'break'

    def find_selection_event(self, event=None) -> str:
        """Perform find selection operation."""
        SearchDialog.find_selection(self)
        return 'break'

    def replace_event(self, event=None) -> str:
        """Perform replace operation."""
        ReplaceDialog.replace(self)
        return 'break'

    def find_all(self, s):
        """Highlight all occurrences of string s, and select the first one.

        If the string has already been highlighted, jump to the next
        occurrence after the current selection. (You cannot go backwards
        using the button, but you can manually place the cursor anywhere
        in the document to start searching from that point.)
        """
        if hasattr(self, '_highlight') and self._highlight == s:
            try:
                if self.get(tkinter.SEL_FIRST, tkinter.SEL_LAST) == s:
                    return self.find_selection_event(None)
                # user must have changed the selection
                found = self.tag_nextrange('found', tkinter.SEL_LAST)
            except tkinter.TclError:
                # user must have unset the selection
                found = self.tag_nextrange('found', tkinter.INSERT)
            if not found:
                # at last occurrence, scroll back to the top
                found = self.tag_nextrange('found', 1.0)
            if found:
                self.do_highlight(found[0], found[1])
        else:
            # find all occurrences of string s;
            # adapted from O'Reilly's Python in a Nutshell
            # remove previous uses of tag 'found', if any
            self.tag_remove('found', '1.0', tkinter.END)
            if s:
                self._highlight = s
                # start from the beginning (and when we come to the end, stop)
                idx = '1.0'
                while True:
                    # find next occurrence, exit loop if no more
                    idx = self.search(s, idx, nocase=1, stopindex=tkinter.END)
                    if not idx:
                        break
                    # index right after the end of the occurrence
                    lastidx = f'{idx}+{len(s)}c'
                    # tag the whole occurrence (start included, stop excluded)
                    self.tag_add('found', idx, lastidx)
                    # prepare to search for next occurrence
                    idx = lastidx
                # use a red foreground for all the tagged occurrences
                self.tag_config('found', foreground='red')
                found = self.tag_nextrange('found', 1.0)
                if found:
                    self.do_highlight(found[0], found[1])
        return None

    def do_highlight(self, start, end) -> None:
        """Select and show the text from index start to index end."""
        self.see(start)
        self.tag_remove(tkinter.SEL, '1.0', tkinter.END)
        self.tag_add(tkinter.SEL, start, end)
        self.focus_set()

    def goto_line_event(self, event) -> str | None:
        """Perform goto line operation."""
        lineno = simpledialog.askinteger('Goto', 'Go to line number:',
                                         parent=self)
        if lineno is None:
            return 'break'
        if lineno <= 0:
            self.bell()
            return 'break'
        self.mark_set('insert', f'{lineno}.0')
        self.see('insert')
        return None


class EditBoxWindow(Frame):

    """Edit box window."""

    def __init__(self, parent=None, **kwargs) -> None:
        """Initializer."""
        for module in (idlelib, tkinter):
            if isinstance(module, ImportError):
                raise module

        if parent is None:
            # create a new window
            parent = tkinter.Tk()
        self.parent = parent
        super().__init__(parent)
        self.editbox = MultiCallCreator(TextEditor)(self, **kwargs)
        self.editbox.pack(side=tkinter.TOP)
        self.editbox.add_bindings()
        self.bind('<<open-config-dialog>>', self.config_dialog)

        bottom = tkinter.Frame(parent)
        # lower left subframe with a textfield and a Search button
        bottom_left_frame = tkinter.Frame(bottom)
        self.textfield = tkinter.Entry(bottom_left_frame)
        self.textfield.pack(side=tkinter.LEFT, fill=tkinter.X, expand=1)

        button_search = tkinter.Button(bottom_left_frame, text='Find next',
                                       command=self.find)
        button_search.pack(side=tkinter.RIGHT)
        bottom_left_frame.pack(side=tkinter.LEFT, expand=1)

        # lower right subframe which will contain OK and Cancel buttons
        bottom_right_frame = tkinter.Frame(bottom)

        button_ok = tkinter.Button(bottom_right_frame, text='OK',
                                   command=self.pressedOK)
        button_cancel = tkinter.Button(bottom_right_frame, text='Cancel',
                                       command=parent.destroy)
        button_ok.pack(side=tkinter.LEFT, fill=tkinter.X)
        button_cancel.pack(side=tkinter.RIGHT, fill=tkinter.X)
        bottom_right_frame.pack(side=tkinter.RIGHT, expand=1)

        bottom.pack(side=tkinter.TOP)

        # create a toplevel menu
        menubar = tkinter.Menu(self.parent)

        findmenu = tkinter.Menu(menubar)
        findmenu.add_command(label='Find',
                             command=self.editbox.find_event,
                             accelerator='Ctrl+F',
                             underline=0)
        findmenu.add_command(label='Find again',
                             command=self.editbox.find_again_event,
                             accelerator='Ctrl+G',
                             underline=6)
        findmenu.add_command(label='Find all',
                             command=self.find_all,
                             underline=5)
        findmenu.add_command(label='Find selection',
                             command=self.editbox.find_selection_event,
                             accelerator='Ctrl+F3',
                             underline=5)
        findmenu.add_command(label='Replace',
                             command=self.editbox.replace_event,
                             accelerator='Ctrl+H',
                             underline=0)
        menubar.add_cascade(label='Find', menu=findmenu, underline=0)

        editmenu = tkinter.Menu(menubar)
        editmenu.add_command(label='Cut',
                             command=self.editbox.cut,
                             accelerator='Ctrl+X',
                             underline=2)
        editmenu.add_command(label='Copy',
                             command=self.editbox.copy,
                             accelerator='Ctrl+C',
                             underline=0)
        editmenu.add_command(label='Paste',
                             command=self.editbox.paste,
                             accelerator='Ctrl+V',
                             underline=0)
        editmenu.add_separator()
        editmenu.add_command(label='Select all',
                             command=self.editbox.select_all,
                             accelerator='Ctrl+A',
                             underline=7)
        editmenu.add_command(label='Clear selection',
                             command=self.editbox.remove_selection,
                             accelerator='Esc')
        menubar.add_cascade(label='Edit', menu=editmenu, underline=0)

        optmenu = tkinter.Menu(menubar)
        optmenu.add_command(label='Settings...',
                            command=self.config_dialog,
                            underline=0)
        menubar.add_cascade(label='Options', menu=optmenu, underline=0)

        # display the menu
        self.parent.config(menu=menubar)
        self.pack()

    def edit(self, text: str, jumpIndex: int | None = None,  # noqa: N803
             highlight: str | None = None) -> str | None:
        """Provide user with editor to modify text.

        :param text: the text to be edited
        :param jumpIndex: position at which to put the caret
        :param highlight: each occurrence of this substring will be
            highlighted
        :return: the modified text, or None if the user didn't save the
            text file in his text editor
        """
        self.text = None
        # put given text into our textarea
        self.editbox.insert(tkinter.END, text)
        # wait for user to push a button which will destroy (close) the window
        # enable word wrap
        self.editbox.tag_add('all', '1.0', tkinter.END)
        self.editbox.tag_config('all', wrap=tkinter.WORD)
        # start search if required
        if highlight:
            self.find_all(highlight)
        if jumpIndex:
            # lines are indexed starting at 1
            line = text[:jumpIndex].count('\n') + 1
            column = jumpIndex - (text[:jumpIndex].rfind('\n') + 1)
            # don't know how to place the caret, but scrolling to the right
            # line should already be helpful.
            self.editbox.see(f'{line}.{column}')
        # wait for user to push a button which will destroy (close) the window
        self.parent.mainloop()
        return self.text

    def find_all(self, target) -> None:
        """Perform find all operation."""
        self.textfield.insert(tkinter.END, target)
        self.editbox.find_all(target)

    def find(self) -> None:
        """Perform find operation."""
        # get text to search for
        s = self.textfield.get()
        if s:
            self.editbox.find_all(s)

    def config_dialog(self, event=None) -> None:
        """Show config dialog."""
        ConfigDialog(self, 'Settings')

    def pressedOK(self) -> None:  # noqa: N802
        """Perform OK operation.

        Called when user pushes the OK button. Saves the buffer into a
        variable, and closes the window.
        """
        self.text = self.editbox.get('1.0', tkinter.END)
        self.parent.destroy()

    def debug(self, event=None) -> str:
        """Call quit() and return 'break'."""
        self.quit()
        return 'break'


class Tkdialog:

    """The dialog window for image info."""

    def __init__(self, photo_description, photo, filename) -> None:
        """Initializer."""
        for module in (idlelib, tkinter):
            if isinstance(module, ImportError):
                raise module

        self.root = tkinter.Tk()
        # "%dx%d%+d%+d" % (width, height, xoffset, yoffset)
        self.root.geometry(f'{int(pywikibot.config.tkhorsize)}x'
                           f'{int(pywikibot.config.tkvertsize)}+10-10')

        self.root.title(filename)
        self.photo_description = photo_description
        self.filename = filename
        self.photo = photo
        self.skip = False
        self.exit = False

        # --Init of the widgets
        # The image
        self.image = self.get_image(self.photo, 800, 600)
        self.image_panel = tkinter.Label(self.root, image=self.image)

        self.image_panel.image = self.image

        # The filename
        self.filename_label = tkinter.Label(self.root,
                                            text='Suggested filename')
        self.filename_field = tkinter.Entry(self.root, width=100)
        self.filename_field.insert(tkinter.END, filename)

        # The description
        self.description_label = tkinter.Label(self.root,
                                               text='Suggested description')
        self.description_scrollbar = tkinter.Scrollbar(self.root,
                                                       orient=tkinter.VERTICAL)
        self.description_field = tkinter.Text(self.root)
        self.description_field.insert(tkinter.END, photo_description)
        self.description_field.config(
            state=tkinter.NORMAL, height=12, width=100, padx=0, pady=0,
            wrap=tkinter.WORD, yscrollcommand=self.description_scrollbar.set)
        self.description_scrollbar.config(command=self.description_field.yview)

        # The buttons
        self.ok_button = tkinter.Button(self.root, text='OK',
                                        command=self.ok_file)
        self.skip_button = tkinter.Button(self.root, text='Skip',
                                          command=self.skip_file)

        # --Start grid

        # The image
        self.image_panel.grid(row=0, column=0, rowspan=11, columnspan=4)

        # The buttons
        self.ok_button.grid(row=11, column=1, rowspan=2)
        self.skip_button.grid(row=11, column=2, rowspan=2)

        # The filename
        self.filename_label.grid(row=13, column=0)
        self.filename_field.grid(row=13, column=1, columnspan=3)

        # The description
        self.description_label.grid(row=14, column=0)
        self.description_field.grid(row=14, column=1, columnspan=3)
        self.description_scrollbar.grid(row=14, column=5)

    @staticmethod
    def get_image(photo, width, height):
        """Take the BytesIO object and build an imageTK thumbnail."""
        try:
            from PIL import Image, ImageTk
        except ImportError:
            pywikibot.warning('This script requires ImageTk from the'
                              'Python Imaging Library (PIL).')
            raise

        image = Image.open(photo)
        image.thumbnail((width, height))
        return ImageTk.PhotoImage(image)

    def ok_file(self) -> None:
        """The user pressed the OK button."""
        self.filename = self.filename_field.get()
        self.photo_description = self.description_field.get(0.0, tkinter.END)
        self.root.destroy()

    def skip_file(self) -> None:
        """The user pressed the Skip button."""
        self.skip = True
        self.root.destroy()

    def show_dialog(self) -> tuple[str, str, bool]:
        """Activate the dialog.

        :return: new description, name, and if the image is skipped
        """
        self.root.mainloop()
        return self.photo_description, self.filename, self.skip



"""This module can do slight modifications to tidy a wiki page's source code.

The changes are not supposed to change the look of the rendered wiki page.

If you wish to run this as an stand-alone script, use::

    scripts/cosmetic_changes.py

For regular use, it is recommended to put this line into your user config::

    cosmetic_changes = True

You may enable cosmetic changes for additional languages by adding the
dictionary cosmetic_changes_enable to your user-config.py. It should contain
a tuple of languages for each site where you wish to enable in addition to
your own langlanguage if cosmetic_changes_mylang_only is True (see below).
Please set your dictionary by adding such lines to your user config::

    cosmetic_changes_enable['wikipedia'] = ('de', 'en', 'fr')

There is another config variable: You can set::

    cosmetic_changes_mylang_only = False

if you're running a bot on multiple sites and want to do cosmetic changes on
all of them, but be careful if you do.

You may disable cosmetic changes by adding the all unwanted languages to
the `dictionary cosmetic_changes_disable` in your user config file
(`user-config.py`). It should contain a tuple of languages for each site
where you wish to disable cosmetic changes. You may use it with
`cosmetic_changes_mylang_only` is False, but you can also disable your
own language. This also overrides the settings in the dictionary
`cosmetic_changes_enable`. Please set this dictionary by adding such
lines to your user config file::

    cosmetic_changes_disable['wikipedia'] = ('de', 'en', 'fr')

You may disable cosmetic changes for a given script by appending the all
unwanted scripts to the list cosmetic_changes_deny_script in your
user-config.py. By default it contains cosmetic_changes.py itself and touch.py.
This overrides all other enabling settings for cosmetic changes. Please modify
the given list by adding such lines to your user-config.py::

    cosmetic_changes_deny_script.append('your_script_name_1')

or by adding a list to the given one::

    cosmetic_changes_deny_script += ['your_script_name_1',
                                     'your_script_name_2']
"""
#
# (C) Pywikibot team, 2006-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import re
from contextlib import suppress
from enum import IntEnum
from typing import Any, cast
from urllib.parse import urlparse, urlunparse

import pywikibot
from pywikibot import exceptions, i18n, textlib
from pywikibot.backports import Callable, Match, Pattern
from pywikibot.site import Namespace
from pywikibot.tools import first_lower, first_upper
from pywikibot.tools.chars import url2string
from pywikibot.userinterfaces.transliteration import NON_ASCII_DIGITS


try:
    import stdnum.isbn as stdnum_isbn
except ImportError:
    stdnum_isbn = None


# Subpage templates. Must be in lower case,
# whereas subpage itself must be case sensitive
# This is also used by interwiki.py
# TODO: Maybe move it to family file and implement global instances
moved_links = {
    'ar': (['documentation', 'template documentation', 'شرح', 'توثيق'],
           '/doc'),
    'bn': ('documentation', '/doc'),
    'ca': ('ús de la plantilla', '/ús'),
    'cs': ('dokumentace', '/doc'),
    'da': ('dokumentation', '/doc'),
    'de': ('dokumentation', '/Meta'),
    'dsb': (['dokumentacija', 'doc'], '/Dokumentacija'),
    'en': (['documentation', 'template documentation', 'template doc',
            'doc', 'documentation, template'], '/doc'),
    'es': (['documentación', 'documentación de plantilla'], '/doc'),
    'eu': ('txantiloi dokumentazioa', '/dok'),
    'fa': (['documentation', 'template documentation', 'template doc',
            'doc', 'توضیحات', 'زیرصفحه توضیحات'], '/doc'),
    # fi: no idea how to handle this type of subpage at :Metasivu:
    'fi': ('mallineohje', None),
    'fr': (['/documentation', 'documentation', 'doc_modèle',
            'documentation modèle', 'documentation modèle compliqué',
            'documentation modèle en sous-page',
            'documentation modèle compliqué en sous-page',
            'documentation modèle utilisant les parserfunctions en sous-page',
            ],
           '/Documentation'),
    'hsb': (['dokumentacija', 'doc'], '/Dokumentacija'),
    'hu': ('sablondokumentáció', '/doc'),
    'id': ('template doc', '/doc'),
    'ilo': ('documentation', '/doc'),
    'ja': ('documentation', '/doc'),
    'ka': ('თარგის ინფო', '/ინფო'),
    'ko': ('documentation', '/설명문서'),
    'ms': ('documentation', '/doc'),
    'no': ('dokumentasjon', '/dok'),
    'nn': ('dokumentasjon', '/dok'),
    'pl': ('dokumentacja', '/opis'),
    'pt': (['documentação', '/doc'], '/doc'),
    'ro': ('documentaţie', '/doc'),
    'ru': ('doc', '/doc'),
    'simple': (['documentation',
                'template documentation',
                'template doc',
                'doc',
                'documentation, template'], '/doc'),
    'sk': ('dokumentácia', '/Dokumentácia'),
    'sv': ('dokumentation', '/dok'),
    'uk': (['документація', 'doc', 'documentation'], '/Документація'),
    'ur': (['دستاویز', 'توثيق', 'شرح', 'توضیحات',
            'documentation', 'template doc', 'doc',
            'documentation, template'], '/doc'),
    'vi': ('documentation', '/doc'),
    'zh': (['documentation', 'doc'], '/doc'),
}

# Template which should be replaced or removed.
# Use a list with two entries. The first entry will be replaced by the second.
# Examples:
# For removing {{Foo}}, the list must be:
#           ('Foo', None),
#
# The following also works:
#           ('Foo', ''),
#
# For replacing {{Foo}} with {{Bar}} the list must be:
#           ('Foo', 'Bar'),
#
# This also removes all template parameters of {{Foo}}
# For replacing {{Foo}} with {{Bar}} but keep the template
# parameters in its original order, please use:
#           ('Foo', 'Bar\\g<parameters>'),

deprecatedTemplates = {
    'wikipedia': {
        'de': [
            ('Belege', 'Belege fehlen\\g<parameters>'),
            ('Quelle', 'Belege fehlen\\g<parameters>'),
            ('Quellen', 'Belege fehlen\\g<parameters>'),
            ('Quellen fehlen', 'Belege fehlen\\g<parameters>'),
        ],
        'ur': [
            ('Infobox former country',
             'خانہ معلومات سابقہ ملک\\g<parameters>'),
            ('Infobox Former Country',
             'خانہ معلومات سابقہ ملک\\g<parameters>'),
        ],
    }
}

main_sortkey = {
    '_default': ' ',
    'ar': '*',
}
"""Sort key to specify the main article within a category.

The sort key must be one of ``' '``, ``'!'``, ``'*'``, ``'#'`` and is
used like a pipe link but sorts the page in front of the alphabetical
order. This dict is used in
:meth:`CosmeticChangesToolkit.standardizePageFooter`.

.. versionadded:: 9.3
"""


class CANCEL(IntEnum):

    """Cancel level to ignore exceptions.

    If an error occurred and either skips the page or the method
    or a single match. ALL raises the exception.

    .. versionadded:: 6.3
    """

    ALL = 0
    PAGE = 1
    METHOD = 2
    MATCH = 3


def _format_isbn_match(match: Match[str], *, strict: bool = True) -> str:
    """Helper function to validate and format a single matched ISBN."""
    if not stdnum_isbn:
        raise NotImplementedError(
            'ISBN functionality not available. Install stdnum package.')

    isbn = match['code']
    try:
        stdnum_isbn.validate(isbn)
    except stdnum_isbn.ValidationError as e:
        if strict:
            raise
        pywikibot.log(f'ISBN "{isbn}" validation error: {e}')
        return isbn

    return stdnum_isbn.format(isbn)


def _reformat_ISBNs(text: str, *, strict: bool = True) -> str:
    """Helper function to normalise ISBNs in text.

    :raises Exception: Invalid ISBN encountered when strict enabled
    """
    return textlib.reformat_ISBNs(
        text, lambda match: _format_isbn_match(match, strict=strict))


class CosmeticChangesToolkit:

    """Cosmetic changes toolkit.

    .. versionchanged:: 7.0
       `from_page()` method was removed
    """

    def __init__(self, page: pywikibot.page.BasePage, *,
                 show_diff: bool = False,
                 ignore: IntEnum = CANCEL.ALL) -> None:
        """Initializer.

        .. versionchanged:: 5.2
           instantiate the CosmeticChangesToolkit from a page object;
           only allow keyword arguments except for page parameter;
           `namespace` and `pageTitle` parameters are deprecated

        .. versionchanged:: 7.0
           `namespace` and `pageTitle` parameters were removed

        :param page: the Page object containing the text to be modified
        :param show_diff: show difference after replacements
        :param ignore: ignores if an error occurred and either skips the page
            or only that method. It can be set one of the CANCEL constants
        """
        self.site = page.site
        self.title = page.title()
        self.namespace = page.namespace()

        self.show_diff = show_diff
        self.template = self.namespace == Namespace.TEMPLATE
        self.talkpage = self.namespace >= 0 and self.namespace % 2 == 1
        self.ignore = ignore

        self.common_methods = [
            self.commonsfiledesc,
            self.fixSelfInterwiki,
            self.standardizePageFooter,
            self.fixSyntaxSave,
            self.cleanUpLinks,
            self.cleanUpSectionHeaders,
            self.putSpacesInLists,
            self.translateAndCapitalizeNamespaces,
            self.translateMagicWords,
            self.replaceDeprecatedTemplates,
            self.resolveHtmlEntities,
            self.removeEmptySections,
            self.removeUselessSpaces,
            self.removeNonBreakingSpaceBeforePercent,

            self.fixHtml,
            self.fixReferences,
            self.fixStyle,
            self.fixTypo,

            self.fixArabicLetters,
        ]
        if stdnum_isbn:
            self.common_methods.append(self.fix_ISBN)

    def safe_execute(self, method: Callable[[str], str], text: str) -> str:
        """Execute the method and catch exceptions if enabled."""
        result = None
        try:
            result = method(text)
        except Exception as e:
            if self.ignore != CANCEL.METHOD:
                raise

            pywikibot.warning(
                f'Unable to perform "{method.__name__}" on "{self.title}"!')
            pywikibot.error(e)

        return text if result is None else result

    def _change(self, text: str) -> str:
        """Execute all clean up methods."""
        for method in self.common_methods:
            text = self.safe_execute(method, text)
        return text

    def change(self, text: str) -> bool | str:
        """Execute all clean up methods and catch errors if activated."""
        try:
            new_text = self._change(text)
        except Exception as e:
            if self.ignore == CANCEL.PAGE:
                pywikibot.warning(
                    f'Skipped "{self.title}", because an error occurred.'
                )
                pywikibot.error(e)
                return False
            raise

        if self.show_diff:
            pywikibot.showDiff(text, new_text)
        return new_text

    def fixSelfInterwiki(self, text: str) -> str:
        """Interwiki links to the site itself are displayed like local links.

        Remove their language code prefix.
        """
        if not self.talkpage and pywikibot.calledModuleName() != 'interwiki':
            interwikiR = re.compile(
                rf'\[\[(?: *:)? *{self.site.code} *: *([^\[\]\n]*)\]\]'
            )
            text = interwikiR.sub(r'[[\1]]', text)
        return text

    def standardizePageFooter(self, text: str) -> str:
        """Standardize page footer.

        Makes sure that interwiki links and categories are put into the
        correct position and into the right order.

        The page footer consists of the following parts in that sequence:

        1. categories
        2. additional information depending on the local site policy
        3. interwiki

        .. versionchanged:: 9.3
           uses :attr:`main_sortkey` to determine the sort key for the
           main article within a category. If the main article has a
           sort key already, it will not be changed any longer.

        :param text: text to be modified
        :return: the modified *text*
        :raises ValueError: wrong value of sortkey in
            :attr:`main_sortkey` for the given site
        """
        categories = []
        interwiki_links = {}

        # get categories
        if not self.template:
            categories = textlib.getCategoryLinks(text, site=self.site)

        if not self.talkpage:
            subpage = False
            if self.template:
                loc = i18n.translate(self.site.code, moved_links)
                if loc is not None and loc[1] in self.title:
                    subpage = True

            # get interwiki
            interwiki_links = textlib.getLanguageLinks(
                text, insite=self.site, template_subpage=subpage)

            # remove interwiki
            text = textlib.removeLanguageLinks(text, site=self.site)

        # add categories, main to top
        if categories:
            # TODO: Sort categories in alphabetic order, e.g. using
            # categories.sort()? (T100265)
            # TODO: Get main categories from Wikidata?
            main = pywikibot.Category(self.site, 'Category:' + self.title)
            if main in categories:
                main = categories.pop(categories.index(main))
                if main.sortKey:
                    sortkey = main.sortKey
                else:
                    sortkey = i18n.translate(self.site, main_sortkey,
                                             fallback=['_default'])
                    if sortkey not in [' ', '*', '!', '#']:
                        raise ValueError(
                            f'sort key for {self.site} is {sortkey} but must'
                            "be one of ' ', '*', '!', '#'")
                main = pywikibot.Category(main, sort_key=sortkey)
                categories.insert(0, main)

            text = textlib.replaceCategoryLinks(text, categories,
                                                site=self.site)

        # add interwiki
        if interwiki_links:
            text = textlib.replaceLanguageLinks(text, interwiki_links,
                                                site=self.site,
                                                template=self.template,
                                                template_subpage=subpage)

        return text

    def translateAndCapitalizeNamespaces(self, text: str) -> str:
        """Use localized namespace names.

        .. versionchanged:: 7.4
           No longer expect a specific namespace alias for File:
        """
        # arz uses English stylish codes
        if self.site.sitename == 'wikipedia:arz':
            return text
        # wiki links aren't parsed here.
        exceptions = ['nowiki', 'comment', 'math', 'pre']

        for namespace in self.site.namespaces.values():
            if namespace == Namespace.MAIN:
                # skip main (article) namespace
                continue
            # a clone is needed. Won't change the namespace dict
            namespaces = list(namespace)
            if (namespace == Namespace.FILE
                    and self.site.family.name == 'wikipedia'):
                if self.site.code in ('en', 'fr'):
                    # do not change "Image" on en-wiki and fr-wiki
                    with suppress(ValueError):
                        namespaces.remove('Image')
                if self.site.code == 'hu':
                    # do not change "Kép" on hu-wiki
                    with suppress(ValueError):
                        namespaces.remove('Kép')
                elif self.site.code == 'pt':
                    # use "Imagem" by default on pt-wiki (per T57242)
                    with suppress(ValueError):
                        namespaces.insert(
                            0, namespaces.pop(namespaces.index('Imagem')))
            # final namespace variant
            final_ns = namespaces.pop(0)
            if namespace in (Namespace.USER, Namespace.USER_TALK):
                # skip localized user namespace, maybe gender is used
                namespaces = ['User'
                              if namespace == Namespace.USER
                              else 'User talk']
            # lowerspaced and underscored namespaces
            for i, item in enumerate(namespaces):
                ns = item.replace(' ', '[ _]')
                namespaces[i] = f'[{ns[0]}{ns[0].lower()}]{ns[1:]}'
            namespaces.append(first_lower(final_ns))
            if final_ns and namespaces:
                if (self.site.sitename == 'wikipedia:pt'
                        and namespace == Namespace.FILE):
                    # only change on these file extensions (per T57242)
                    extensions = ('png', 'gif', 'jpg', 'jpeg', 'svg', 'tiff',
                                  'tif')
                    text = textlib.replaceExcept(
                        text,
                        r'\[\[\s*({}) *:(?P<name>[^\|\]]*?\.({}))'
                        r'(?P<label>.*?)\]\]'
                        .format('|'.join(namespaces), '|'.join(extensions)),
                        fr'[[{final_ns}:\g<name>\g<label>]]',
                        exceptions)
                else:
                    text = textlib.replaceExcept(
                        text,
                        r'\[\[\s*({}) *:(?P<nameAndLabel>.*?)\]\]'
                        .format('|'.join(namespaces)),
                        fr'[[{final_ns}:\g<nameAndLabel>]]',
                        exceptions)
        return text

    def translateMagicWords(self, text: str) -> str:
        """Use localized magic words."""
        # not wanted at ru
        # arz uses English stylish codes
        # no need to run on English wikis
        if self.site.code in ['arz', 'en', 'ru']:
            return text

        def init_cache() -> None:
            for magicword in ('img_thumbnail', 'img_left', 'img_center',
                              'img_right', 'img_none', 'img_framed',
                              'img_frameless', 'img_border', 'img_upright',
                              'img_baseline', 'img_sub', 'img_super',
                              'img_top', 'img_text_top', 'img_middle',
                              'img_bottom', 'img_text_bottom'):
                aliases = self.site.getmagicwords(magicword)
                if len(aliases) > 1:
                    cache.update((alias, aliases[0]) for alias in aliases[1:]
                                 if '$1' not in alias)
            if not cache:
                cache[False] = True  # signal there is nothing to replace

        def replace_magicword(match: Match[str]) -> str:
            """Replace magic words in file link params, leaving captions."""
            linktext = match.group()
            if cache.get(False):
                return linktext

            params = match.group(2)  # includes pre-leading |
            if not params:
                return linktext

            if not cache:
                init_cache()

            # do the magic job
            marker = textlib.findmarker(params)
            params = textlib.replaceExcept(
                params, r'\|', marker, ['link', 'template'])
            parts = params.split(marker)
            replaced = '|'.join(cache.get(p.strip(), p) for p in parts)

            # extract namespace
            m = cast(Match[str],
                     re.match(r'\[\[\s*(?P<namespace>[^:]+)\s*:', linktext))

            return f'[[{m["namespace"]}:{match["filename"]}{replaced}]]'

        cache: dict[bool | str, Any] = {}
        exceptions = ['comment', 'nowiki', 'pre', 'syntaxhighlight']
        regex = textlib.get_regexes('file', self.site)[0]
        return textlib.replaceExcept(
            text, regex, replace_magicword, exceptions)

    def cleanUpLinks(self, text: str) -> str:
        """Tidy up wikilinks found in a string.

        This function will:

        * Replace underscores with spaces
        * Move leading and trailing spaces out of the wikilink and into
          the surrounding text
        * Convert URL-encoded characters into Unicode-encoded characters
        * Move trailing characters out of the link and make the link
          without using a pipe, if possible
        * Capitalize the article title of the link, if appropriate

        .. versionchanged:: 8.4
           Convert URL-encoded characters if a link is an interwiki link
           or different from main namespace.

        :param text: string to perform the clean-up on
        :return: text with tidied wikilinks
        """
        # helper function which works on one link and either returns it
        # unmodified, or returns a replacement.
        def handleOneLink(match: Match[str]) -> str:
            # Convert URL-encoded characters to str
            titleWithSection = url2string(match['titleWithSection'],
                                          encodings=self.site.encodings())
            label = match['label']
            trailingChars = match['linktrail']
            newline = match['newline'] or ''
            # entire link but convert URL-encoded text
            oldlink = url2string(match.group(),
                                 encodings=self.site.encodings())

            is_interwiki = None
            with suppress(exceptions.ServerError):
                is_interwiki = self.site.isInterwikiLink(titleWithSection)
            if is_interwiki is not False:
                return oldlink

            # The link looks like this:
            # [[page_title|link_text]]trailing_chars
            # We only work on namespace 0 because pipes and linktrails work
            # differently for images and categories.
            page = pywikibot.Page(pywikibot.Link(titleWithSection, self.site))
            in_main_namespace = None
            with suppress(exceptions.InvalidTitleError):
                in_main_namespace = page.namespace() == Namespace.MAIN
            if not in_main_namespace:
                return oldlink

            # Replace underlines by spaces, remove double spaces
            titleWithSection = re.sub('[_ ]+', ' ', titleWithSection)
            # Remove unnecessary leading spaces from title,
            # but remember if we did this because we eventually want
            # to re-add it outside of the link later.
            titleLength = len(titleWithSection)
            titleWithSection = titleWithSection.lstrip()
            hadLeadingSpaces = not newline \
                and len(titleWithSection) != titleLength
            hadTrailingSpaces = False
            # Remove unnecessary trailing spaces from title,
            # but remember if we did this because it may affect
            # the linktrail and because we eventually want to
            # re-add it outside of the link later.
            if not trailingChars:
                titleLength = len(titleWithSection)
                titleWithSection = titleWithSection.rstrip()
                hadTrailingSpaces = len(titleWithSection) != titleLength

            if not titleWithSection:
                # just skip empty links.
                return match.group()

            # Remove unnecessary initial and final spaces from label.
            # Please note that some editors prefer spaces around pipes.
            # (See [[en:Wikipedia:Semi-bots]]). We remove them anyway.
            if label is None:
                label = titleWithSection
            else:
                # Remove unnecessary leading spaces from label,
                # but remember if we did this because we want
                # to re-add it outside of the link later.
                labelLength = len(label)
                label = label.lstrip()
                hadLeadingSpaces = not newline and len(label) != labelLength
                # Remove unnecessary trailing spaces from label,
                # but remember if we did this because it affects
                # the linktrail.
                if not trailingChars:
                    labelLength = len(label)
                    label = label.rstrip()
                    hadTrailingSpaces = len(label) != labelLength

            if trailingChars:
                label += trailingChars

            if self.site.siteinfo['case'] == 'first-letter':
                firstcase_title = first_lower(titleWithSection)
                firstcase_label = first_lower(label)
            else:
                firstcase_title = titleWithSection
                firstcase_label = label

            if firstcase_label == firstcase_title:
                newLink = f'[[{label}]]'
            # Check if we can create a link with trailing characters
            # instead of a pipelink
            elif (firstcase_label.startswith(firstcase_title)
                  and trailR.sub('', label[len(titleWithSection):]) == ''):
                newLink = (f'[[{label[:len(titleWithSection)]}]]'
                           f'{label[len(titleWithSection):]}')
            else:
                # Try to capitalize the first letter of the title.
                # Not useful for languages that don't capitalize nouns.
                # TODO: Add a configuration variable for each site,
                # which determines if the link target is written in
                # uppercase
                if self.site.sitename == 'wikipedia:de':
                    titleWithSection = first_upper(titleWithSection)
                newLink = f'[[{titleWithSection}|{label}]]'

            # re-add spaces that were pulled out of the link.
            # Examples:
            #   text[[ title ]]text        -> text [[title]] text
            #   text[[ title | name ]]text -> text [[title|name]] text
            #   text[[ title |name]]text   -> text[[title|name]]text
            #   text[[title| name]]text    -> text [[title|name]]text
            return f"{newline}{' ' if hadLeadingSpaces else ''}" \
                   f"{newLink}{' ' if hadTrailingSpaces else ''}"

        trailR = re.compile(self.site.linktrail())
    # The regular expression which finds links. Results consist of four groups:
    # group <newline> depends whether the links starts with a new line.
    # group <titleWithSection> is the page title and section, that is,
    # everything before | or ]. It'll include the # to make life easier for us.
    # group <label> is the alternative link title between | and ].
    # group <linktrail> is the link trail after ]] which are part of the word.
    # note that the definition of 'letter' varies from language to language.
        linkR = re.compile(
            r'(?P<newline>[\n]*)\[\[(?P<titleWithSection>[^\]\|]+)'
            r'(\|(?P<label>[^\]\|]*))?\]\](?P<linktrail>'
            + self.site.linktrail() + ')')

        return textlib.replaceExcept(text, linkR, handleOneLink,
                                     ['comment', 'math', 'nowiki', 'pre',
                                      'startspace'])

    def resolveHtmlEntities(self, text: str) -> str:
        """Replace HTML entities with string."""
        ignore = [
            38,     # Ampersand (&amp;)
            39,     # Single quotation mark (&quot;) per T26093
            60,     # Less than (&lt;)
            62,     # Greater than (&gt;)
            91,     # Opening square bracket ([)
                    # - sometimes used intentionally inside links
            93,     # Closing square bracket (])
                    # - used intentionally inside links
            124,    # Vertical bar (|)
                    # - used intentionally in navigation bar templates on w:de
            160,    # Non-breaking space (&nbsp;)
                    # - not supported by Firefox textareas
            173,    # Soft-hyphen (&shy;) - enable editing
            8206,   # Left-to-right mark (&ltr;)
            8207,   # Right-to-left mark (&rtl;)
        ]
        if self.template:
            ignore.append(32)  # Space ( )
            ignore.append(58)  # Colon (:)
        # TODO: T254350 - what other extension tags should be avoided?
        # (graph, math, score, timeline, etc.)
        return pywikibot.html2unicode(
            text, ignore=ignore, exceptions=['comment', 'syntaxhighlight'])

    def removeEmptySections(self, text: str) -> str:
        """Cleanup empty sections."""
        # userspace contains article stubs without nobots/in use templates
        if self.namespace == Namespace.USER:
            return text

        skippings = ['comment', 'category']
        skip_regexes = textlib.get_regexes(skippings, self.site)
        # site defined templates
        skip_templates = {
            'cs': ('Pahýl[ _]část',),  # stub section
        }
        if self.site.code in skip_templates:
            for template in skip_templates[self.site.code]:
                skip_regexes.append(
                    re.compile(r'\{\{\s*%s\s*\}\}' % template, re.IGNORECASE))
        # empty lists
        skip_regexes.append(re.compile(r'(?m)^[\*#] *$'))

        # get stripped sections
        stripped_text = textlib.removeLanguageLinks(text, self.site, '\n')
        for reg in skip_regexes:
            stripped_text = reg.sub(r'', stripped_text)

        strip_sections = textlib.extract_sections(
            stripped_text, self.site).sections

        # get proper sections
        header, sections, footer = textlib.extract_sections(text, self.site)

        if len(strip_sections) > len(sections):
            # there must be something wrong with the extracted sections; skip
            return text

        # iterate stripped sections and create a new page body
        new_body: textlib.SectionList = textlib.SectionList()
        for i, strip_section in enumerate(strip_sections):
            current_dep = sections[i].level
            try:
                next_dep = sections[i + 1].level
            except IndexError:
                next_dep = 0

            if strip_section.content.strip() or current_dep < next_dep:
                new_body.extend(sections[i])
        return header + ''.join(new_body) + footer

    def removeUselessSpaces(self, text: str) -> str:
        """Cleanup multiple or trailing spaces."""
        exceptions = ['comment', 'math', 'nowiki', 'pre', 'syntaxhighlight',
                      'startspace', 'table']
        if self.site.sitename != 'wikipedia:cs':
            exceptions.append('template')

        return textlib.replaceExcept(text, r'(?m)[\t ]+( |$)', r'\1',
                                     exceptions, site=self.site)

    def removeNonBreakingSpaceBeforePercent(self, text: str) -> str:
        """Remove a non-breaking space between number and percent sign.

        Newer MediaWiki versions automatically place a non-breaking
        space in front of a percent sign, so it is no longer required to
        place it manually.
        """
        return textlib.replaceExcept(
            text, r'(\d)&(?:nbsp|#160|#x[Aa]0);%', r'\1 %', ['timeline'])

    def cleanUpSectionHeaders(self, text: str) -> str:
        """Add a space between the equal signs and the section title.

        Example::

            ==Section title==

        becomes::

        == Section title ==

        .. note:: This space is recommended in the syntax help on the
           English and German Wikipedias. It is not wanted on Lojban and
           English Wiktionaries (:phab:`T168399`, :phab:`T169064`) and
           it might be that it is not wanted on other wikis. If there
           are any complaints, please file a bug report.
        """
        if self.site.sitename in ['wiktionary:jbo', 'wiktionary:en']:
            return text
        return textlib.replaceExcept(
            text,
            r'(?m)^(={1,6})[ \t]*(?P<title>.*[^\s=])[ \t]*\1[ \t]*\r?\n',
            r'\1 \g<title> \1\n',
            ['comment', 'math', 'nowiki', 'pre'])

    def putSpacesInLists(self, text: str) -> str:
        """Add a space between the * or # and the text.

        .. note:: This space is recommended in the syntax help on the
           English, German and French Wikipedias. It might be that it
           is not wanted on other wikis. If there are any complaints,
           please file a bug report.
        """
        if not self.template:
            exceptions = ['comment', 'math', 'nowiki', 'pre',
                          'syntaxhighlight', 'template', 'timeline',
                          self.site.redirect_regex]
            text = textlib.replaceExcept(
                text,
                r'(?m)'
                r'^(?P<bullet>[:;]*(\*+|#+)[:;\*#]*)(?P<char>[^\s\*#:;].+?)',
                r'\g<bullet> \g<char>',
                exceptions)
        return text

    def replaceDeprecatedTemplates(self, text: str) -> str:
        """Replace deprecated templates."""
        exceptions = ['comment', 'math', 'nowiki', 'pre']
        builder = textlib.MultiTemplateMatchBuilder(self.site)

        if self.site.family.name in deprecatedTemplates \
           and self.site.code in deprecatedTemplates[self.site.family.name]:
            for template in deprecatedTemplates[
                    self.site.family.name][self.site.code]:
                old, new = template
                new = '{{%s}}' % new if new else ''

                text = textlib.replaceExcept(
                    text,
                    builder.pattern(old),
                    new, exceptions)

        return text

    # from fixes.py
    def fixSyntaxSave(self, text: str) -> str:
        """Convert weblinks to wikilink, fix link syntax."""
        def replace_link(match: Match[str]) -> str:
            """Create a string to replace a single link."""
            replacement = '[['
            if re.match(
                    r'(?:{}):'.format('|'.join(
                        (*self.site.namespaces[Namespace.FILE],
                         *self.site.namespaces[Namespace.CATEGORY]))),
                    match['link']):
                replacement += ':'

            replacement += match['link']
            if match['title']:
                replacement += '|' + match['title']

            return replacement + ']]'

        exceptions = ['comment', 'math', 'nowiki', 'pre', 'startspace',
                      'syntaxhighlight']

        # link to the wiki working on
        # Only use suffixes for article paths
        for suffix in self.site._interwiki_urls(True):
            http_url = self.site.base_url(suffix, 'http')
            if self.site.protocol() == 'http':
                https_url = None
            else:
                https_url = self.site.base_url(suffix, 'https')

            # compare strings without the protocol, if they are empty support
            # also no prefix (//en.wikipedia.org/…)
            http = urlparse(http_url)
            https = urlparse(https_url)
            if https_url is not None and http.netloc == https.netloc:
                urls = ['(?:https?:)?'
                        + re.escape(urlunparse(('', *http[1:])))]
            else:
                urls = [re.escape(url) for url in (http_url, https_url)
                        if url is not None]

            for url in urls:
                # unescape {} placeholder
                url = url.replace(r'\{\}', '{title}')

                # Only include links which don't include the separator
                # as the wikilink won't support additional parameters
                separator = '?&' if '?' in suffix else '?'

                # Match first a non space in the title to prevent that multiple
                # spaces at the end without title will be matched by it
                title_regex = (rf'(?P<link>[^{separator}]+?)'
                               r'(\s+(?P<title>[^\s].*?))')
                url_regex = fr'\[\[?{url}?\s*\]\]?'
                text = textlib.replaceExcept(
                    text,
                    url_regex.format(title=title_regex),
                    replace_link, exceptions, site=self.site)

        # external link in/starting with double brackets
        text = textlib.replaceExcept(
            text,
            r'\[\[(?P<url>https?://[^\]]+?)\]\]?',
            r'[\g<url>]', exceptions, site=self.site)

        # external link and description separated by a pipe, with
        # whitespace in front of the pipe, so that it is clear that
        # the dash is not a legitimate part of the URL.
        text = textlib.replaceExcept(
            text,
            r'\[(?P<url>https?://[^\|\] \r\n]+?) +\| *(?P<label>[^\|\]]+?)\]',
            r'[\g<url> \g<label>]', exceptions)

        # dash in external link, where the correct end of the URL can
        # be detected from the file extension. It is very unlikely that
        # this will cause mistakes.
        extensions = [fr'\.{ext}'
                      for ext in ['pdf', 'html?', 'php', 'aspx?', 'jsp']]

        return textlib.replaceExcept(
            text,
            r'\[(?P<url>https?://[^\|\] ]+?(' + '|'.join(extensions) + r')) *'
            r'\| *(?P<label>[^\|\]]+?)\]',
            r'[\g<url> \g<label>]', exceptions
        )

    def fixHtml(self, text: str) -> str:
        """Replace html markups with wikitext markups."""
        def replace_header(match: Match[str]) -> str:
            """Create a header string for replacing."""
            depth = int(match[1])
            return r'{0} {1} {0}'.format('=' * depth, match[2])

        # Everything case-insensitive (?i)
        # Keep in mind that MediaWiki automatically converts <br> to <br />
        exceptions = ['comment', 'math', 'nowiki', 'pre', 'startspace',
                      'syntaxhighlight']
        text = textlib.replaceExcept(text, r'(?i)<(b|strong)>(.*?)</\1>',
                                     r"'''\2'''", exceptions, site=self.site)
        text = textlib.replaceExcept(text, r'(?i)<(i|em)>(.*?)</\1>',
                                     r"''\2''", exceptions, site=self.site)
        # horizontal line without attributes in a single line
        text = textlib.replaceExcept(text, r'(?i)([\r\n])<hr[ /]*>([\r\n])',
                                     r'\1----\2', exceptions)
        # horizontal line with attributes; can't be done with wiki syntax
        # so we only make it XHTML compliant
        text = textlib.replaceExcept(text, r'(?i)<hr ([^>/]+?)>',
                                     r'<hr \1 />',
                                     exceptions)
        # a header where only spaces are in the same line
        text = textlib.replaceExcept(
            text,
            r'(?i)(?<=[\r\n]) *<h([1-7])> *([^<]+?) *</h\1> *(?=[\r\n])',
            replace_header,
            exceptions)
        # TODO: maybe we can make the bot replace <p> tags with \r\n's.
        return text

    def fixReferences(self, text: str) -> str:
        """Fix references tags."""
        # See also
        # https://en.wikipedia.org/wiki/User:AnomieBOT/source/tasks/OrphanReferenceFixer.pm
        exceptions = ['comment', 'math', 'nowiki', 'pre', 'syntaxhighlight',
                      'startspace']

        # it should be name = " or name=" NOT name   ="
        text = re.sub(r'(?i)<ref +name(= *| *=)"', r'<ref name="', text)
        # remove empty <ref/>-tag
        text = textlib.replaceExcept(text,
                                     r'(?i)(<ref\s*/>|<ref *>\s*</ref>)',
                                     r'', exceptions)
        text = textlib.replaceExcept(text,
                                     r'(?i)<ref\s+([^>]+?)\s*>\s*</ref>',
                                     r'<ref \1/>', exceptions)
        return text

    def fixStyle(self, text: str) -> str:
        """Convert prettytable to wikitable class."""
        exceptions = ['comment', 'math', 'nowiki', 'pre', 'startspace',
                      'syntaxhighlight']
        if self.site.code in ('de', 'en'):
            text = textlib.replaceExcept(text,
                                         r'(class="[^"]*)prettytable([^"]*")',
                                         r'\1wikitable\2', exceptions)
        return text

    def fixTypo(self, text: str) -> str:
        """Fix units."""
        exceptions: list[str | Pattern[str]] = [
            'comment',
            'gallery',
            'hyperlink',
            'interwiki',
            'link',
            'nowiki',
            'math',
            'pre',
            'startspace',
            'syntaxhighlight',
        ]

        # change <number> ccm -> <number> cm³
        text = textlib.replaceExcept(text, r'(\d)\s*(?:&nbsp;)?ccm',
                                     r'\1&nbsp;cm³', exceptions,
                                     site=self.site)
        # Solve wrong Nº sign with °C or °F
        # additional exception requested on fr-wiki for this stuff
        pattern = re.compile('«.*?»')
        exceptions.append(pattern)
        text = textlib.replaceExcept(text, r'(\d)\s*(?:&nbsp;)?[º°]([CF])',
                                     r'\1&nbsp;°\2', exceptions,
                                     site=self.site)
        text = textlib.replaceExcept(text, 'º([CF])', '°' + r'\1',
                                     exceptions,
                                     site=self.site)
        return text

    def fixArabicLetters(self, text: str) -> str:
        """Fix Arabic and Persian letters."""
        if self.site.code not in ['ckb', 'fa']:
            return text

        exceptions: list[str | Pattern[str]] = [
            'file',
            'gallery',
            'hyperlink',
            'interwiki',
            'inputbox',
            # FIXME: but changes letters inside wikilinks
            # 'link',
            'math',
            'pre',
            'template',
            'timeline',
            'ref',
            'startspace',
            'syntaxhighlight',
        ]

        digits = NON_ASCII_DIGITS['fa']
        faChrs = 'ءاآأإئؤبپتثجچحخدذرزژسشصضطظعغفقکگلمنوهیةيك' + digits

        # not to let bot edits in ascii numerals content
        exceptions.append(re.compile(f'[^{faChrs}] *?"*? *?, *?[^{faChrs}]'))
        text = textlib.replaceExcept(text, ',', '،', exceptions,
                                     site=self.site)
        if self.site.code == 'ckb':
            text = textlib.replaceExcept(text,
                                         '\u0647([.\u060c_<\\]\\s])',
                                         '\u06d5\\1', exceptions,
                                         site=self.site)
            text = textlib.replaceExcept(text, 'ه\u200c', 'ە', exceptions,
                                         site=self.site)
            text = textlib.replaceExcept(text, 'ه', 'ھ', exceptions,
                                         site=self.site)
        text = textlib.replaceExcept(text, 'ك', 'ک', exceptions,
                                     site=self.site)
        text = textlib.replaceExcept(text, '[ىي]', 'ی', exceptions,
                                     site=self.site)

        return text

    def commonsfiledesc(self, text: str) -> str:
        """Clean up file descriptions on Wikimedia Commons.

        It works according to [1] and works only on pages in the file
        namespace on Wikimedia Commons.

        [1]:
        https://commons.wikimedia.org/wiki/Commons:Tools/pywiki_file_description_cleanup
        """
        if (self.site.sitename != 'commons:commons'
                or self.namespace == Namespace.FILE):
            return text

        # section headers to {{int:}} versions
        exceptions = ['comment', 'includeonly', 'math', 'noinclude', 'nowiki',
                      'pre', 'syntaxhighlight', 'ref', 'timeline']
        text = textlib.replaceExcept(text,
                                     r'([\r\n]|^)\=\= *Summary *\=\=',
                                     r'\1== {{int:filedesc}} ==',
                                     exceptions, True)
        text = textlib.replaceExcept(
            text,
            r'([\r\n])\=\= *\[\[Commons:Copyright tags\|Licensing\]\]: *\=\=',
            r'\1== {{int:license-header}} ==', exceptions, True)
        text = textlib.replaceExcept(
            text,
            r'([\r\n])'
            r'\=\= *(Licensing|License information|{{int:license}}) *\=\=',
            r'\1== {{int:license-header}} ==', exceptions, True)

        # frequent field values to {{int:}} versions
        text = textlib.replaceExcept(
            text,
            r'([\r\n]\|[Ss]ource *\= *)'
            r'(?:[Oo]wn work by uploader|[Oo]wn work|[Ee]igene [Aa]rbeit) *'
            r'([\r\n])',
            r'\1{{own}}\2', exceptions, True)
        text = textlib.replaceExcept(
            text,
            r'(\| *Permission *\=) *(?:[Ss]ee below|[Ss]iehe unten) *([\r\n])',
            r'\1\2', exceptions, True)

        # added to transwikied pages
        text = textlib.replaceExcept(text, r'__NOTOC__', '', exceptions, True)

        # tracker element for js upload form
        text = textlib.replaceExcept(
            text,
            r'<!-- *{{ImageUpload\|(?:full|basic)}} *-->',
            '', exceptions[1:], True)
        text = textlib.replaceExcept(text, r'{{ImageUpload\|(?:basic|full)}}',
                                     '', exceptions, True)

        # duplicated section headers
        text = textlib.replaceExcept(
            text,
            r'([\r\n]|^)\=\= *{{int:filedesc}} *\=\=(?:[\r\n ]*)\=\= *'
            r'{{int:filedesc}} *\=\=',
            r'\1== {{int:filedesc}} ==', exceptions, True)
        text = textlib.replaceExcept(
            text,
            r'([\r\n]|^)\=\= *{{int:license-header}} *\=\=(?:[\r\n ]*)'
            r'\=\= *{{int:license-header}} *\=\=',
            r'\1== {{int:license-header}} ==', exceptions, True)
        return text

    def fix_ISBN(self, text: str) -> str:
        """Hyphenate ISBN numbers."""
        return _reformat_ISBNs(text, strict=self.ignore != CANCEL.MATCH)



#!/usr/bin/env python3
"""Wrapper script to invoke pywikibot-based scripts.

This wrapper script invokes script by its name in this search order:

1. Scripts listed in `user_script_paths` list inside your user config
   settings file (usually `user-config.py`) in the given order. Refer
   :ref:`External Script Path Settings`.
2. User scripts residing in `scripts/userscripts` (directory mode only).
3. Scripts residing in `scripts` folder (directory mode only).
4. Maintenance scripts residing in `scripts/maintenance` (directory mode only).
5. Site-package scripts (site-package only)
6. Framework scripts residing in `pywikibot/scripts`.

This wrapper script is able to invoke scripts even if the script name is
misspelled. In directory mode it also checks package dependencies.

Run scripts with pywikibot in directory mode using::

    python pwb.py <pwb options> <name_of_script> <options>

or run scripts with pywikibot installed as a site package using::

    pwb <pwb options> <name_of_script> <options>

This wrapper script uses the package directory to store all user files,
will fix up search paths so the package does not need to be installed, etc.

Currently, `<pwb options>` are :ref:`global options`. This can be used
for tests to set the default site (see :phab:`T216825`)::

    python pwb.py -lang:de bot_tests -v

.. seealso:: :mod:`pwb` entry point
.. versionchanged:: 7.0
   pwb wrapper was added to the Python site package lib.
.. versionchanged:: 7.7
   pwb wrapper is able to set ``PYWIKIBOT_TEST_...`` environment variables,
   see :ref:`Environment variables`.
.. versionchanged:: 8.0
   renamed to wrapper.py.
.. versionchanged:: 9.4
   enable external scripts via entry points.
"""
#
# (C) Pywikibot team, 2012-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import importlib.metadata
import os
import sys
import types
from difflib import get_close_matches
from importlib import import_module
from pathlib import Path
from time import sleep
from warnings import warn


pwb = None
site_package = False


def check_pwb_versions(package: str) -> None:
    """Validate package version and scripts version.

    Rules:
        - Pywikibot version must not be older than scrips version
        - Scripts version must not be older than previous Pywikibot version
          due to deprecation policy
    """
    from packaging.version import Version

    scripts_version = Version(getattr(package, '__version__', pwb.__version__))
    wikibot_version = Version(pwb.__version__)

    if scripts_version.release > wikibot_version.release:
        print(f'WARNING: Pywikibot version {wikibot_version} is behind '
              f'scripts package version {scripts_version}.\n'
              'Your Pywikibot may need an update or be misconfigured.\n')

    # calculate previous minor release
    if wikibot_version.minor > 0:
        prev_wikibot = Version(
            f'{wikibot_version.major}.{wikibot_version.minor - 1}.'
            f'{wikibot_version.micro}'
        )

        if scripts_version.release < prev_wikibot.release:
            print(f'WARNING: Scripts package version {scripts_version} is '
                  f'behind legacy Pywikibot version {prev_wikibot} and '
                  f'current version {wikibot_version}\n'
                  'Your scripts may need an update or be misconfigured.\n')
    elif scripts_version.release < wikibot_version.release:
        print(f'WARNING: Scripts package version {scripts_version} is behind '
              f'current version {wikibot_version}\n'
              'Your scripts may need an update or be misconfigured.\n')

    del Version


# The following snippet was developed by Ned Batchelder (and others)
# for coverage [1], with Python 3 support [2] added later,
# and is available under the BSD license (see [3])
# [1]
# https://bitbucket.org/ned/coveragepy/src/b5abcee50dbe/coverage/execfile.py
# [2]
# https://bitbucket.org/ned/coveragepy/src/fd5363090034/coverage/execfile.py
# [3]
# https://bitbucket.org/ned/coveragepy/src/2c5fb3a8b81c/setup.py?at=default#cl-31


def run_python_file(filename: str, args: list[str], package=None) -> None:
    """Run a python file as if it were the main program on the command line.

    .. versionchanged:: 7.7
       Set and restore ``PYWIKIBOT_TEST_...`` environment variables.

    :param filename: The path to the file to execute, it need not be a
        .py file.
    :param args: is the argument list to present as sys.argv, as strings.
    :param package: The package of the script. Used for checks.
    :type package: Optional[module]
    """
    # Create a module to serve as __main__
    old_main_mod = sys.modules['__main__']
    main_mod = types.ModuleType('__main__')
    sys.modules['__main__'] = main_mod
    main_mod.__file__ = filename
    main_mod.__builtins__ = sys.modules['builtins']
    if package:
        main_mod.__package__ = package.__name__
        check_pwb_versions(package)

    # Set sys.argv and the first path element properly.
    old_argv = sys.argv
    old_argvu = pwb.argvu

    # set environment values
    old_env = os.environ.copy()
    for key, value in environ:
        os.environ[key] = value

    sys.argv = [filename, *args]
    pwb.argvu = [Path(filename).stem, *args]
    sys.path.insert(0, os.path.dirname(filename))

    try:
        with open(filename, 'rb') as f:
            source = f.read()
        exec(compile(source, filename, 'exec', dont_inherit=True),
             main_mod.__dict__)
    finally:
        # Restore the old __main__
        sys.modules['__main__'] = old_main_mod

        # Restore the old argv and path
        sys.argv = old_argv
        sys.path.pop(0)
        pwb.argvu = old_argvu

# end of snippet from coverage

        # Restore environment values
        for key, value in environ:
            if key in old_env:
                os.environ[key] = old_env[key]
            else:
                del os.environ[key]


def handle_args(
    _,
    *args: str,
) -> tuple[str, list[str], list[str], list[str]]:
    """Handle args and get filename.

    .. versionchanged:: 7.7
       Catch ``PYWIKIBOT_TEST_...`` environment variables.

    :return: filename, script args, local pwb args, environment variables
    """
    fname = None
    local = []
    env = []
    for index, arg in enumerate(args, start=1):
        if arg in ('-version', '--version'):
            fname = 'version.py'
        elif arg in ('pwb', 'pwb.py', 'wrapper', 'wrapper.py'):
            pass
        elif arg.startswith('-'):
            local.append(arg)
        elif arg.startswith('PYWIKIBOT_TEST_'):
            var, _, val = arg.partition('=')
            env.append((var, val or '1'))
        else:
            fname = arg
            if not fname.endswith('.py'):
                fname += '.py'
        if fname:
            break
    else:
        index = 0

    return fname, list(args[index:]), local, env


def _print_requirements(requirements,
                        script,
                        variant) -> None:
    """Print pip command to install requirements."""
    if not requirements:
        return

    if len(requirements) > 1:
        format_string = '\nPackages necessary for {} are {}.'
    else:
        format_string = '\nA package necessary for {} is {}.'
    print(format_string.format(script or 'pywikibot', variant))
    print('Please update required module{} with:\n\n'
          .format('s' if len(requirements) > 1 else ''))

    for requirement in requirements:
        print(f"    pip install \"{str(requirement).partition(';')[0]}\"\n")


def check_modules(script: str | None = None) -> bool:
    """Check whether mandatory modules are present.

    This also checks Python version when importing dependencies from
    setup.py

    :param script: The script name to be checked for dependencies
    :return: True if all dependencies are installed
    :raise RuntimeError: wrong Python version found in setup.py
    """
    from packaging.requirements import Requirement

    from setup import script_deps

    missing_requirements = []
    version_conflicts = []

    if script:
        dependencies = script_deps.get(Path(script).name, [])
    else:
        from setup import dependencies

    for dependency in dependencies:
        if dependency.startswith('packaging'):
            # Ignore these dependencies because ImportError is raised in an
            # early state when they are imported in backports. They are already
            # used at this point. This is a workaround for toolforge where some
            # modules are not installed as a site-package.
            # TODO: Check imports from external source
            continue

        requirement = Requirement(dependency)
        if requirement.marker is None or requirement.marker.evaluate():
            try:
                instlld_vrsn = importlib.metadata.version(requirement.name)
            except importlib.metadata.PackageNotFoundError as e:
                missing_requirements.append(requirement)
                print(e)
            else:
                if instlld_vrsn not in requirement.specifier:
                    version_conflicts.append(requirement)
                    print(
                        f'{requirement.name} version {instlld_vrsn} is '
                        f'installed but {requirement.specifier} is required'
                    )

    del Requirement
    del script_deps

    _print_requirements(missing_requirements, script, 'missing')
    _print_requirements(version_conflicts, script, 'outdated')

    if version_conflicts and not missing_requirements:  # pragma: no cover
        print('\nYou may continue on your own risk; type CTRL-C to stop.')
        try:
            sleep(5)
        except KeyboardInterrupt:
            return False

    return not missing_requirements


filename, script_args, global_args, environ = handle_args(*sys.argv)

# Search for user config file (user-config.py) before creating one.
# If successful, user config file already exists in one of the candidate
# directories. See config.py for details on search order.
# Use env var to communicate to config.py pwb.py location (bug T74918).
wrapper_dir = Path(__file__).parent
os.environ['PYWIKIBOT_DIR_PWB'] = str(wrapper_dir)

try:
    import pywikibot as pwb
except RuntimeError as e:  # pragma: no cover
    os.environ['PYWIKIBOT_NO_USER_CONFIG'] = '2'
    import pywikibot as pwb

    # user config file to be created
    if filename is not None and not (filename.startswith('generate_')
                                     or filename == 'version.py'):
        from pywikibot.config import user_config_file
        if user_config_file != 'user-config.py':
            # do not create a user config file if name is not default
            sys.exit(e)

        print('NOTE: user-config.py was not found!')
        print('Please follow the prompts to create it:')
        run_python_file(
            str(wrapper_dir.joinpath('generate_user_files.py')), [])
        # because we have loaded pywikibot without user-config.py loaded,
        # we need to re-start the entire process. Ask the user to do so.
        print('Now, you have to re-execute the command to start your script.')
        sys.exit(1)
except ModuleNotFoundError as module:  # raised in textlib or backports
    print(f'\n{module.msg}\nPlease install it with\n\n'
          f'    pip install {module.name}')
    sys.exit()


def find_alternates(filename, script_paths):
    """Search for similar filenames in the given script paths."""
    from pywikibot import config, error, info, input_choice, warning
    from pywikibot.bot import QuitKeyboardInterrupt, ShowingListOption

    assert config.pwb_close_matches > 0, \
        'config.pwb_close_matches must be greater than 0'
    assert 0.0 < config.pwb_cut_off < 1.0, \
        'config.pwb_cut_off must be a float in range [0, 1]'

    error(f'{filename} not found! Misspelling?')

    scripts = {}

    for folder in script_paths:
        if not folder.exists():
            warning(
                f'{folder} does not exists; remove it from user_script_paths')
            continue
        for script_name in folder.iterdir():
            name, suffix = script_name.stem, script_name.suffix
            if suffix == '.py' and not name.startswith('__'):
                scripts[name] = script_name

    # remove .py for better matching
    filename = filename[:-3]
    similar_scripts = get_close_matches(filename, scripts,
                                        config.pwb_close_matches,
                                        config.pwb_cut_off)
    if not similar_scripts:
        return None

    if len(similar_scripts) == 1:
        script = similar_scripts[0]
        wait_time = config.pwb_autostart_waittime
        info('NOTE: Starting the most similar script '
             f'<<lightyellow>>{script}.py<<default>>\n'
             f'      in {wait_time} seconds; type CTRL-C to stop.')
        try:
            sleep(wait_time)  # Wait a bit to let it be cancelled
        except KeyboardInterrupt:
            return None
    else:
        msg = '\nThe most similar scripts are:'
        alternatives = ShowingListOption(similar_scripts, pre=msg, post='')
        try:
            _, script = input_choice('Which script to be run:',
                                     alternatives, default='1')
        except QuitKeyboardInterrupt:
            return None
        print()
    return str(scripts[script])


def find_filename(filename):
    """Search for the filename in the given script paths.

    .. versionchanged:: 7.0
       Search users_scripts_paths in config.base_dir
    .. versionchanged:: 9.0
       Add config.base_dir to search path
    .. versionchanged:: 9.4
       Search in entry point paths
    """
    from pywikibot import config
    path_list = []  # paths to find misspellings

    def test_paths(paths, root: Path):
        """Search for filename in given paths within 'root' base directory."""
        for file_package in paths:
            package = file_package.split('.')
            path = [*package, filename]
            testpath = root.joinpath(*path)
            if testpath.exists():
                return str(testpath)
            path_list.append(testpath.parent)
        return None

    # search through user scripts paths
    user_script_paths = ['']
    if config.user_script_paths:
        if isinstance(config.user_script_paths, list):
            user_script_paths += config.user_script_paths
        else:
            warn("'user_script_paths' must be a list,\n"
                 f'found: {type(config.user_script_paths).__name__}.'
                 ' Ignoring this setting.', stacklevel=2)

    found = test_paths(user_script_paths, Path(config.base_dir))
    if found:
        return found

    if site_package:  # pragma: no cover
        # search for entry points
        import importlib
        from importlib.metadata import entry_points

        from pywikibot.i18n import set_messages_package

        if sys.version_info < (3, 10):
            entry_points_items = [
                ep for ep in entry_points().get('pywikibot', [])
                if ep.name == 'scriptspath'
            ]
        else:
            entry_points_items = entry_points(
                name='scriptspath',
                group='pywikibot',
            )

        for ep in entry_points_items:
            path = ep.load()
            found = test_paths([''], path)
            if found:
                i18n_package = path.stem + '.i18n'
                if importlib.import_module(i18n_package).__file__ is not None:
                    set_messages_package(i18n_package)
                return found

    else:  # search in scripts folder
        script_paths = [
            'scripts.userscripts',
            'scripts',
            'scripts.maintenance',
        ]
        found = test_paths(script_paths, wrapper_dir.parents[1])
        if found:
            return found

    # search for system scripts in pywikibot.scripts directory
    found = test_paths([''], wrapper_dir)
    if found:
        return found

    return find_alternates(filename, path_list)


def execute() -> bool:
    """Parse arguments, extract filename and run the script.

    .. versionadded:: 7.0
       renamed from :func:`main`
    """
    global filename

    # If -nolog options are in global args, add them to
    # script args.
    if '-nolog' in global_args:
        script_args.append('-nolog')

    if global_args:  # don't use sys.argv
        unknown_args = pwb.handle_args(global_args)
        if unknown_args:
            print('ERROR: unknown pwb.py argument{}: {}\n'
                  .format('' if len(unknown_args) == 1 else 's',
                          ', '.join(unknown_args)))
            return False

    if not filename:
        return False

    file_package = None

    if not os.path.exists(filename):
        filename = find_filename(filename)
        if filename is None:
            return True

    # When both pwb.py and the filename to run are within the current
    # working directory:
    # a) set __package__ as if called using python -m scripts.blah.foo
    # b) set __file__ to be relative, so it can be relative in backtraces,
    #    and __file__ *appears* to be an unstable path to load data from.
    # This is a rough (and quick!) emulation of 'package name' detection.
    # a much more detailed implementation is in coverage's find_module.
    # https://bitbucket.org/ned/coveragepy/src/default/coverage/execfile.py
    cwd = Path.cwd()
    syspath = Path(sys.argv[0])
    absolute_path = syspath.parent
    file_path = Path(filename)
    if absolute_path == cwd and cwd in file_path.parents:
        relative_filename = file_path.relative_to(absolute_path)
        # remove the filename, and use '.' instead of path separator.
        file_package = str(relative_filename.parent).replace(os.sep, '.')
        filename = os.path.join(os.curdir, str(relative_filename))

    module = None
    if file_package:
        try:
            module = sys.modules[file_package]
        except KeyError:
            try:
                module = import_module(file_package)
            except ImportError as e:
                warn(f'Parent module {file_package} not found: {e}',
                     ImportWarning, stacklevel=2)

    help_option = any(arg.startswith('-help:') or arg == '-help'
                      for arg in script_args)
    if site_package or check_modules(filename) or help_option:
        run_python_file(filename, script_args, module)
    return True


def main() -> None:
    """Script entry point. Print doc if necessary.

    .. versionchanged:: 7.0
       previous implementation was renamed to :func:`execute`
    """
    if not check_modules():
        sys.exit()

    if not execute():
        print(__doc__)


def run() -> None:  # pragma: no cover
    """Site package entry point. Print doc if necessary.

    .. versionadded:: 7.0
    """
    global site_package
    site_package = True
    if not execute():
        print(__doc__)


if __name__ == '__main__':
    main()



"""Library module representing MediaWiki sites (wikis)."""
#
# (C) Pywikibot team, 2021-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations


# to prevent ImportError exception due to circular imports
from pywikibot.site._namespace import Namespace, NamespacesDict  # isort: split

from pywikibot.site._apisite import APISite
from pywikibot.site._basesite import BaseSite
from pywikibot.site._datasite import DataSite
from pywikibot.site._namespace import NamespaceArgType  # noqa: F401
from pywikibot.site._obsoletesites import ClosedSite, RemovedSite
from pywikibot.site._siteinfo import Siteinfo
from pywikibot.site._tokenwallet import TokenWallet


__all__ = ('APISite', 'BaseSite', 'ClosedSite', 'DataSite', 'RemovedSite',
           'Namespace', 'NamespacesDict', 'Siteinfo', 'TokenWallet')

# iiprop file information to get, used in several places
_IIPROP = (
    'timestamp', 'user', 'comment', 'url', 'size', 'sha1', 'mime', 'mediatype',
    'archivename', 'bitdepth',
)



"""Special bot library containing UploadRobot.

Do not import classes directly from here but from specialbots.
"""
#
# (C) Pywikibot team, 2003-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import os
import tempfile
from contextlib import suppress
from http import HTTPStatus
from pathlib import Path
from urllib.parse import urlparse

import requests

import pywikibot
from pywikibot import config
from pywikibot.backports import Callable
from pywikibot.bot import BaseBot, QuitKeyboardInterrupt
from pywikibot.comms import http
from pywikibot.exceptions import APIError, FatalServerError, NoPageError


class UploadRobot(BaseBot):

    """Upload bot."""

    post_processor: Callable[[str, str | None], None] | None = None
    """If this attribute is set to a callable, the :meth:`run` method calls it
    after upload. The parameters passed to the callable is the origin
    *file_url* passed to the :meth:`upload` method and the *filename* returned
    from that method. It can be used like this:

    .. code-block:: python

       def summarize(old: str, new: str | None) -> None:
           if new is None:
               print(f'{old} was ignored')
           else:
               print(f'{old} was uploaded as {new}')

       bot = UploadRobot('Myfile.bmp')
       bot.post_processor = summarize
       bot.run()

    .. versionadded:: 9.1
    """

    def __init__(self, url: list[str] | str, *,
                 url_encoding=None,
                 description: str = '',
                 use_filename=None,
                 keep_filename: bool = False,
                 verify_description: bool = True,
                 ignore_warning: bool | list = False,
                 target_site=None,
                 aborts: bool | list | None = None,
                 chunk_size: int = 0,
                 asynchronous: bool = False,
                 summary: str | None = None,
                 filename_prefix: str | None = None,
                 force_if_shared: bool = False,
                 **kwargs) -> None:
        """Initializer.

        .. versionchanged:: 6.2
           asynchronous upload is used if *asynchronous* parameter is set
        .. versionchanged:: 6.4
           *force_if_shared* parameter was added

        :param url: path to url or local file, or list of urls or paths
            to local files.
        :param description: Description of file for its page. If multiple files
            are uploading the same description is used for every file.
        :param use_filename: Specify title of the file's page. If multiple
            files are uploading it asks to change the name for second, third,
            etc. files, otherwise the last file will overwrite the other.
        :param keep_filename: Set to True to keep original names of urls and
            files, otherwise it will ask to enter a name for each file.
        :param summary: Summary of the upload
        :param verify_description: Set to True to proofread the description.
        :param ignore_warning: Set this to True to upload even if another file
            would be overwritten or another mistake would be risked. Set it to
            an array of warning codes to selectively ignore specific warnings.
        :param target_site: Set the site to upload to. If target site is not
            given it's taken from user config file (user-config.py).
        :type target_site: object
        :param aborts: List of the warning types to abort upload on. Set to
            True to abort on any warning.
        :param chunk_size: Upload the file in chunks (more overhead, but
            restartable) specified in bytes. If no value is specified the file
            will be uploaded as whole.
        :param asynchronous: Make potentially large file operations
            asynchronous on the server side when possible.
        :param filename_prefix: Specify prefix for the title of every
            file's page.
        :param force_if_shared: Upload the file even if it's currently
            shared to the target site (e.g. when moving from Commons to another
            wiki)
        :keyword bool always: Disables any input, requires that either
            ignore_warning or aborts are set to True and that the
            description is also set. It overwrites verify_description to
            False and keep_filename to True.
        """
        super().__init__(**kwargs)
        if self.opt.always:
            if ignore_warning is not True and aborts is not True:
                raise ValueError(
                    'When always is set to True, '
                    'ignore_warning or aborts must be set to True.')
            if not description:
                raise ValueError(
                    'When always is set to True, the description must be set.')

        self.url = [url] if isinstance(url, str) else url
        self.url_encoding = url_encoding
        self.description = description
        self.use_filename = use_filename
        self.keep_filename = keep_filename or self.opt.always
        self.verify_description = verify_description and not self.opt.always
        self.ignore_warning = ignore_warning
        self.aborts = aborts or []
        self.chunk_size = chunk_size
        self.asynchronous = asynchronous
        self.summary = summary
        self.filename_prefix = filename_prefix
        self.force_if_shared = force_if_shared

        if config.upload_to_commons:
            default_site = pywikibot.Site('commons')
        else:
            default_site = pywikibot.Site()
        self.target_site = target_site or default_site

    def read_file_content(self, file_url: str):
        """Return name of temp file in which remote file is saved."""
        pywikibot.info('Reading file ' + file_url)

        handle, tempname = tempfile.mkstemp()
        path = Path(tempname)
        size = 0

        dt_gen = (el for el in (15, 30, 45, 60, 120, 180, 240, 300))
        while True:
            file_len = path.stat().st_size
            if file_len:
                pywikibot.info('Download resumed.')
                headers = {'Range': f'bytes={file_len}-'}
            else:
                headers = {}

            with open(path, 'ab') as fd:
                os.lseek(handle, file_len, 0)
                try:
                    response = http.fetch(file_url, stream=True,
                                          headers=headers)
                    response.raise_for_status()

                    # get download info, if available
                    # Note: this is not enough to exclude pages
                    #       e.g. 'application/json' is also not a media
                    if 'text/' in response.headers['Content-Type']:
                        raise FatalServerError('The requested URL was not '
                                               'found on server.')
                    size = max(size,
                               int(response.headers.get('Content-Length', 0)))

                    # stream content to temp file (in chunks of 1Mb)
                    for chunk in response.iter_content(chunk_size=1024 * 1024):
                        fd.write(chunk)

                # raised from connection lost during response.iter_content()
                except requests.ConnectionError:
                    fd.flush()
                    pywikibot.info(
                        f'Connection closed at byte {path.stat().st_size}')
                # raised from response.raise_for_status()
                except requests.HTTPError as e:
                    # exit criteria if size is not available
                    # error on last iteration is OK, we're requesting
                    #    {'Range': 'bytes=file_len-'}
                    err = HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE
                    if response.status_code == err and path.stat().st_size:
                        break
                    raise FatalServerError(str(e)) from e

            if size and size == path.stat().st_size:
                break
            try:
                dt = next(dt_gen)
                pywikibot.info(f'Sleeping for {dt} seconds ...')
                pywikibot.sleep(dt)
            except StopIteration:
                raise FatalServerError('Download failed, too many retries!')

        pywikibot.info(f'Downloaded {path.stat().st_size} bytes')
        return tempname

    def _handle_warning(self, warning: str) -> bool | None:
        """Return whether the warning cause an abort or be ignored.

        :param warning: The warning name
        :return: False if this warning should cause an abort, True if it
            should be ignored or None if this warning has no default
            handler.
        """
        if self.aborts is not True and warning in self.aborts:
            return False
        if self.ignore_warning is True or (self.ignore_warning is not False
                                           and warning in self.ignore_warning):
            return True
        return None if self.aborts is not True else False

    def _handle_warnings(self, warnings):
        messages = '\n'.join(f'{warning.code}: {warning.info}'
                             for warning in sorted(warnings,
                                                   key=lambda w: w.code))
        if len(warnings) > 1:
            messages = '\n' + messages
        pywikibot.info('We got the following warning(s): ' + messages)
        answer = True
        for warning in warnings:
            this_answer = self._handle_warning(warning.code)
            if this_answer is False:
                answer = False
                break
            if this_answer is None:
                answer = None
        if answer is None:
            answer = pywikibot.input_yn('Do you want to ignore?',
                                        default=False, automatic_quit=False)
        return answer

    def process_filename(self, file_url: str) -> str | None:
        """Return base filename portion of *file_url*.

        .. versionchanged:: 10.2
           no longer shows the description if UploadRobot's parameter
           *verify_description* is set to False.

        :param file_url: either a URL or a local file path
        """
        # Isolate the pure name
        filename = file_url
        if '://' in filename:
            # extract the path portion of the URL
            filename = urlparse(filename).path
        filename = os.path.basename(filename)
        if self.use_filename:
            filename = self.use_filename
        if self.filename_prefix:
            filename = self.filename_prefix + filename
        if not self.keep_filename:
            pywikibot.info(f'\nThe filename on the target wiki will default '
                           f'to: {filename}\n')
            assert not self.opt.always
            newfn = pywikibot.input(
                'Enter a better name, or press enter to accept:')
            if newfn != '':
                filename = newfn
        # FIXME: these 2 belong somewhere else, presumably in family
        # forbidden characters are handled by pywikibot/page.py
        forbidden = ':*/\\'  # to be extended
        allowed_formats = self.target_site.file_extensions

        # ask until it's valid
        first_check = True
        while True:
            if not first_check:
                if self.opt.always:
                    filename = None
                else:
                    filename = pywikibot.input('Enter a better name, or press '
                                               'enter to skip the file:')
                if not filename:
                    return None

            first_check = False
            ext = os.path.splitext(filename)[1].lower().strip('.')
            # are any chars in forbidden also in filename?
            invalid = set(forbidden) & set(filename)
            if invalid:
                c = ''.join(invalid)
                pywikibot.info(
                    f'Invalid character(s): {c}. Please try again')
                continue

            if allowed_formats and ext not in allowed_formats:
                if self.opt.always:
                    pywikibot.info('File format is not one of [{}]'
                                   .format(' '.join(allowed_formats)))

                elif pywikibot.input_yn(
                        'File format is not one of [{}], but {!r}. Skip?'
                        .format(' '.join(allowed_formats), ext)):
                    return None

                continue

            potential_file_page = pywikibot.FilePage(self.target_site,
                                                     filename)
            if potential_file_page.exists():
                overwrite = self._handle_warning('exists')
                if overwrite is False:
                    pywikibot.info(
                        'File exists and you asked to abort. Skipping.')
                    return None

                if potential_file_page.has_permission():
                    if overwrite is None:
                        overwrite = not pywikibot.input_yn(
                            f'File with name {filename} already exists.'
                            ' Would you like to change the name?'
                            ' (Otherwise file will be overwritten.)',
                            default=True,
                            automatic_quit=False)
                    if not overwrite:
                        continue
                    break

                pywikibot.info(f'File with name {filename} already exists and '
                               f'cannot be overwritten.')
                continue

            with suppress(NoPageError):
                if (not self.force_if_shared
                        and potential_file_page.file_is_shared()):
                    pywikibot.info(
                        f'File with name {filename} already exists in shared '
                        f'repository and cannot be overwritten.')
                    continue
            break

        # Show the description to verify it for the submission
        if self.description and self.verify_description:
            pywikibot.info(
                f'The suggested description is:\n{self.description}')

        while not self.description or self.verify_description:
            if not self.description:
                pywikibot.info('<<lightred>>It is not possible to upload a '
                               'file without a description.')
            assert not self.opt.always
            # if no description, ask if user want to add one or quit,
            # and loop until one is filled.
            # if self.verify_description, ask if user want to change it
            # or continue.
            if self.description:
                question = 'Do you want to change this description?'
            else:
                question = 'No description was given. Add one?'
            if pywikibot.input_yn(question, default=not self.description,
                                  automatic_quit=self.description):
                from pywikibot import editor as editarticle
                editor = editarticle.TextEditor()
                try:
                    new_description = editor.edit(self.description)
                except ImportError:
                    raise
                except Exception as e:
                    pywikibot.error(e)
                    continue
                # if user saved / didn't press Cancel
                if new_description:
                    self.description = new_description
            elif not self.description:
                raise QuitKeyboardInterrupt
            self.verify_description = False

        return filename

    def abort_on_warn(self, warn_code):
        """Determine if the warning message should cause an abort."""
        return self.aborts is True or warn_code in self.aborts

    def ignore_on_warn(self, warn_code: str):
        """Determine if the warning message should be ignored.

        :param warn_code: The warning message
        """
        return self.ignore_warning is True or warn_code in self.ignore_warning

    def upload_file(self, file_url: str) -> str | None:
        """Upload the image at file_url to the target wiki.

        .. seealso:: :api:`Upload`

        Return the filename that was used to upload the image.
        If the upload fails, ask the user whether to try again or not.
        If the user chooses not to retry, return None.

        .. versionchanged:: 7.0
           If 'copyuploadbaddomain' API error occurred in first step,
           download the file and upload it afterwards

        :param file_url: either a URL or a local file path
        """
        filename = self.process_filename(file_url)
        if not filename:
            return None

        site = self.target_site
        imagepage = pywikibot.FilePage(site, filename)  # normalizes filename
        imagepage.text = self.description

        pywikibot.info(f'Uploading file to {site}...')

        ignore_warnings = self.ignore_warning is True or self._handle_warnings

        download = False
        while True:
            if '://' in file_url \
               and (not site.has_right('upload_by_url') or download):
                try:
                    file_url = self.read_file_content(file_url)
                except FatalServerError as e:
                    pywikibot.error(e)
                    return None

            try:
                success = imagepage.upload(file_url,
                                           ignore_warnings=ignore_warnings,
                                           chunk_size=self.chunk_size,
                                           asynchronous=self.asynchronous,
                                           comment=self.summary)
            except APIError as error:
                if error.code == 'uploaddisabled':
                    pywikibot.error(f'Upload error: Local file uploads are '
                                    f'disabled on {site}.')
                elif error.code == 'copyuploadbaddomain' and not download \
                        and '://' in file_url:
                    pywikibot.error(error)
                    pywikibot.info('Downloading the file and retry...')
                    download = True
                    continue
                else:
                    pywikibot.exception('Upload error: ')
            except Exception:
                pywikibot.exception('Upload error: ')
            else:
                if success:
                    # No warning, upload complete.
                    pywikibot.info(f'Upload of {filename} successful.')
                    self.counter['upload'] += 1
                    return filename  # data['filename']
                pywikibot.info('Upload aborted.')
            break

        return None

    def skip_run(self) -> bool:
        """Check whether processing is to be skipped."""
        # early check that upload is enabled
        if self.target_site.is_uploaddisabled():
            pywikibot.error(f'Upload error: Local file uploads are disabled '
                            f'on {self.target_site}.')
            return True

        # early check that user has proper rights to upload
        self.target_site.login()
        if not self.target_site.has_right('upload'):
            pywikibot.error(f"User '{self.target_site.user()}' does not have "
                            f'upload rights on site {self.target_site}.')
            return True

        return False

    def run(self) -> None:
        """Run bot.

        .. versionchanged:: 9.1
           count uploads.
        """
        if self.skip_run():
            return

        try:
            for file_url in self.url:
                filename = self.upload_file(file_url)
                self.counter['read'] += 1
                if filename:
                    self.counter['upload'] += 1
                if callable(self.post_processor):
                    self.post_processor(file_url, filename)
        except QuitKeyboardInterrupt:
            pywikibot.info(f'\nUser quit {type(self).__name__} bot run...')
        except KeyboardInterrupt:
            if config.verbose_output:
                raise

            pywikibot.info(
                f'\nKeyboardInterrupt during {type(self).__name__} bot run...')
        finally:
            self.exit()
